@@ -11,6 +11,7 @@
 
 #include "data/checksum.h"
 
+#include "init/dev.h"
 #include "init/error.h"
 #include "init/fs.h"
 
@@ -20,6 +21,7 @@
 #include "journal/write.h"
 
 #include "sb/clean.h"
+#include "sb/counters.h"
 
 #include <linux/ioprio.h>
 
@@ -183,98 +185,156 @@ static void journal_buf_realloc(struct journal *j, struct journal_buf *buf)
 	kvfree(new_buf);
 }
 
+static void replicas_refs_put(struct bch_fs *c, darray_replicas_entry_refs *refs)
+{
+	darray_for_each(*refs, i)
+		bch2_replicas_entry_put_many(c, &i->replicas.e, i->nr_refs);
+	refs->nr = 0;
+}
+
+static inline u64 last_uncompleted_write_seq(struct journal *j, u64 seq_completing)
+{
+	u64 seq = journal_last_unwritten_seq(j);
+
+	if (seq <= journal_cur_seq(j) &&
+	    (j->buf[seq & JOURNAL_BUF_MASK].write_done ||
+	     seq == seq_completing))
+		return seq;
+
+	return 0;
+}
+
 static CLOSURE_CALLBACK(journal_write_done)
 {
 	closure_type(w, struct journal_buf, io);
 	struct journal *j = container_of(w, struct journal, buf[w->idx]);
 	struct bch_fs *c = container_of(j, struct bch_fs, journal);
-	u64 seq = le64_to_cpu(w->data->seq);
-	u64 seq_wrote = seq;
+	u64 seq_wrote = le64_to_cpu(w->data->seq);
 	int err = 0;
 
 	bch2_time_stats_update(!JSET_NO_FLUSH(w->data)
 			       ? j->flush_write_time
 			       : j->noflush_write_time, j->write_start_time);
 
-	if (w->had_error) {
-		struct bch_replicas_entry_v1 *r = &journal_seq_pin(j, seq)->devs.e;
+	struct bch_replicas_entry_v1 *r = &journal_seq_pin(j, seq_wrote)->devs.e;
 
-		bch2_devlist_to_replicas(r, BCH_DATA_journal, w->devs_written);
+	if (unlikely(w->failed.nr)) {
+		bch2_replicas_entry_put(c, r);
+		r->nr_devs = 0;
 	}
 
-	if (!w->devs_written.nr)
-		err = bch_err_throw(c, journal_write_err);
+	if (!r->nr_devs && !w->empty) {
+		bch2_devlist_to_replicas(r, BCH_DATA_journal, w->devs_written);
+		err = bch2_replicas_entry_get(c, r);
+		if (err)
+			r->nr_devs = 0;
+	}
 
-	if (err && !bch2_journal_error(j)) {
-		CLASS(printbuf, buf)();
-		bch2_log_msg_start(c, &buf);
+	if (unlikely(w->failed.nr || err)) {
+		CLASS(bch_log_msg, msg)(c);
 
-		if (err == -BCH_ERR_journal_write_err)
-			prt_printf(&buf, "unable to write journal to sufficient devices\n");
-		else
-			prt_printf(&buf, "journal write error marking replicas: %s\n",
-				   bch2_err_str(err));
+		/* Separate ratelimit_states for hard and soft errors */
+		msg.m.suppress = !err
+			? bch2_ratelimit(c)
+			: bch2_ratelimit(c);
 
-		bch2_fs_emergency_read_only2(c, &buf);
+		prt_printf(&msg.m, "error writing journal entry %llu\n", seq_wrote);
+		bch2_io_failures_to_text(&msg.m, c, &w->failed);
 
-		bch2_print_str(c, KERN_ERR, buf.buf);
+		if (!w->devs_written.nr)
+			err = bch_err_throw(c, journal_write_err);
+
+		if (!err) {
+			prt_printf(&msg.m, "wrote degraded to ");
+			bch2_devs_list_to_text(&msg.m, c, &w->devs_written);
+			prt_newline(&msg.m);
+		} else {
+			prt_printf(&msg.m, "error %s\n", bch2_err_str(err));
+			bch2_fs_emergency_read_only2(c, &msg.m);
+		}
 	}
 
 	closure_debug_destroy(cl);
 
+	CLASS(darray_replicas_entry_refs, replicas_refs)();
+
 	spin_lock(&j->lock);
-	BUG_ON(seq < j->pin.front);
-	if (err && (!j->err_seq || seq < j->err_seq))
-		j->err_seq	= seq;
+	BUG_ON(seq_wrote < j->pin.front);
+	if (err && (!j->err_seq || seq_wrote < j->err_seq))
+		j->err_seq = seq_wrote;
 
 	if (!j->free_buf || j->free_buf_size < w->buf_size) {
 		swap(j->free_buf,	w->data);
 		swap(j->free_buf_size,	w->buf_size);
 	}
 
-	if (w->data) {
-		void *buf = w->data;
-		w->data = NULL;
-		w->buf_size = 0;
-
-		spin_unlock(&j->lock);
-		kvfree(buf);
-		spin_lock(&j->lock);
-	}
+	/* kvfree can allocate memory, and can't be called under j->lock */
+	void *buf_to_free __free(kvfree) = w->data;
+	w->data = NULL;
+	w->buf_size = 0;
 
 	bool completed = false;
 	bool last_seq_ondisk_updated = false;
-again:
-	for (seq = journal_last_unwritten_seq(j);
-	     seq <= journal_cur_seq(j);
-	     seq++) {
+
+	u64 seq;
+	while ((seq = last_uncompleted_write_seq(j, seq_wrote))) {
 		w = j->buf + (seq & JOURNAL_BUF_MASK);
-		if (!w->write_done && seq != seq_wrote)
-			break;
 
 		if (!j->err_seq && !w->noflush) {
+			BUG_ON(w->empty && w->last_seq != seq);
+
 			if (j->last_seq_ondisk < w->last_seq) {
-				spin_unlock(&j->lock);
+				bch2_journal_update_last_seq_ondisk(j,
+						w->last_seq + w->empty, &replicas_refs);
 				/*
-				 * this needs to happen _before_ updating
-				 * j->flushed_seq_ondisk, for flushing to work
-				 * properly - when the flush completes replcias
-				 * refs need to have been dropped
-				 * */
-				bch2_journal_update_last_seq_ondisk(j, w->last_seq);
+				 * bch2_journal_update_last_seq_ondisk()
+				 * can return an error if appending to
+				 * replicas_refs failed, but we don't
+				 * care - it's a preallocated darray so
+				 * it'll allways be able to do some
+				 * work, and we have to retry anyways,
+				 * because we have to drop j->lock to
+				 * put the replicas refs before updating
+				 * j->flushed_seq_ondisk
+				 */
+
+				/*
+				 * Do this before updating j->last_seq_ondisk,
+				 * or journal flushing breaks:
+				 */
+				if (replicas_refs.nr) {
+					spin_unlock(&j->lock);
+					replicas_refs_put(c, &replicas_refs);
+					spin_lock(&j->lock);
+					continue;
+				}
+
+				BUG_ON(j->last_seq > j->last_seq);
+				j->last_seq_ondisk = w->last_seq;
 				last_seq_ondisk_updated = true;
-				spin_lock(&j->lock);
-				goto again;
 			}
 
+			/* replicas refs eed to be put first */
 			j->flushed_seq_ondisk = seq;
 		}
 
-		j->seq_ondisk = seq;
-
 		if (w->empty)
 			j->last_empty_seq = seq;
+		j->seq_ondisk = seq;
 
+		closure_wake_up(&w->wait);
+		completed = true;
+	}
+
+	/*
+	 * Writes might complete out of order, but we have to do the completions
+	 * in order: if we complete out of order we note it here so the next
+	 * write completion will pick it up:
+	 */
+	j->buf[seq_wrote & JOURNAL_BUF_MASK].write_done = true;
+	j->pin.front = min(j->pin.back, j->last_seq_ondisk);
+
+	if (completed) {
 		/*
 		 * Updating last_seq_ondisk may let bch2_journal_reclaim_work() discard
 		 * more buckets:
@@ -285,13 +345,6 @@ again:
 		if (j->watermark != BCH_WATERMARK_stripe)
 			journal_reclaim_kick(&c->journal);
 
-		closure_wake_up(&w->wait);
-		completed = true;
-	}
-
-	j->buf[seq_wrote & JOURNAL_BUF_MASK].write_done = true;
-
-	if (completed) {
 		bch2_journal_update_last_seq(j);
 		bch2_journal_space_available(j);
 
@@ -300,8 +353,6 @@ again:
 		journal_wake(j);
 	}
 
-	j->pin.front = min(j->pin.back, j->last_seq_ondisk);
-
 	if (journal_last_unwritten_seq(j) == journal_cur_seq(j) &&
 	    j->reservations.cur_entry_offset < JOURNAL_ENTRY_CLOSED_VAL) {
 		struct journal_buf *buf = journal_cur_buf(j);
@@ -321,12 +372,12 @@ again:
 	 * allocated, in bch2_journal_write() - but the journal write error path
 	 * is special:
 	 */
-	bch2_journal_do_writes(j);
+	bch2_journal_do_writes_locked(j);
 	spin_unlock(&j->lock);
 
 	if (last_seq_ondisk_updated) {
 		bch2_reset_alloc_cursors(c);
-		closure_wake_up(&c->freelist_wait);
+		closure_wake_up(&c->allocator.freelist_wait);
 		bch2_do_discards(c);
 	}
 
@@ -338,22 +389,17 @@ static void journal_write_endio(struct bio *bio)
 	struct journal_bio *jbio = container_of(bio, struct journal_bio, bio);
 	struct bch_dev *ca = jbio->ca;
 	struct journal *j = &ca->fs->journal;
+	struct bch_fs *c = container_of(j, struct bch_fs, journal);
 	struct journal_buf *w = j->buf + jbio->buf_idx;
 
 	bch2_account_io_completion(ca, BCH_MEMBER_ERROR_write,
 				   jbio->submit_time, !bio->bi_status);
 
 	if (bio->bi_status) {
-		bch_err_dev_ratelimited(ca,
-			       "error writing journal entry %llu: %s",
-			       le64_to_cpu(w->data->seq),
-			       bch2_blk_status_to_str(bio->bi_status));
-
-		unsigned long flags;
-		spin_lock_irqsave(&j->err_lock, flags);
+		guard(spinlock_irqsave)(&j->err_lock);
+		bch2_dev_io_failures_mut(&w->failed, ca->dev_idx)->errcode =
+			__bch2_err_throw(c, -blk_status_to_bch_err(bio->bi_status));
 		bch2_dev_list_drop_dev(&w->devs_written, ca->dev_idx);
-		w->had_error = true;
-		spin_unlock_irqrestore(&j->err_lock, flags);
 	}
 
 	closure_put(&w->io);
@@ -367,6 +413,11 @@ static CLOSURE_CALLBACK(journal_write_submit)
 	struct bch_fs *c = container_of(j, struct bch_fs, journal);
 	unsigned sectors = vstruct_sectors(w->data, c->block_bits);
 
+	event_inc_trace(c, journal_write, buf, ({
+		prt_printf(&buf, "seq %llu\n", le64_to_cpu(w->data->seq));
+		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&w->key));
+	}));
+
 	extent_for_each_ptr(bkey_i_to_s_extent(&w->key), ptr) {
 		struct bch_dev *ca = bch2_dev_have_ref(c, ptr->dev);
 
@@ -399,7 +450,6 @@ static CLOSURE_CALLBACK(journal_write_submit)
 
 		bch2_bio_map(bio, w->data, sectors << 9);
 
-		trace_and_count(c, journal_write, bio);
 		closure_bio_submit(bio, cl);
 
 		ja->bucket_seq[ja->cur_idx] = le64_to_cpu(w->data->seq);
@@ -654,13 +704,13 @@ CLOSURE_CALLBACK(bch2_journal_write)
 	closure_type(w, struct journal_buf, io);
 	struct journal *j = container_of(w, struct journal, buf[w->idx]);
 	struct bch_fs *c = container_of(j, struct bch_fs, journal);
-	unsigned nr_rw_members = dev_mask_nr(&c->rw_devs[BCH_DATA_free]);
+	unsigned nr_rw_members = dev_mask_nr(&c->allocator.rw_devs[BCH_DATA_free]);
 	int ret;
 
-	BUG_ON(BCH_SB_CLEAN(c->disk_sb.sb));
 	BUG_ON(!w->write_started);
 	BUG_ON(w->write_allocated);
 	BUG_ON(w->write_done);
+	BUG_ON(journal_last_unallocated_seq(j) != le64_to_cpu(w->data->seq));
 
 	j->write_start_time = local_clock();
 
@@ -713,20 +763,29 @@ CLOSURE_CALLBACK(bch2_journal_write)
 		 * available:
 		 */
 		bch2_journal_space_available(j);
-		bch2_journal_do_writes(j);
+		bch2_journal_do_writes_locked(j);
 	}
 
 	w->devs_written = bch2_bkey_devs(c, bkey_i_to_s_c(&w->key));
 
-	/*
-	 * Mark journal replicas before we submit the write to guarantee
-	 * recovery will find the journal entries after a crash.
-	 */
-	struct bch_replicas_entry_v1 *r = &journal_seq_pin(j, le64_to_cpu(w->data->seq))->devs.e;
-	bch2_devlist_to_replicas(r, BCH_DATA_journal, w->devs_written);
-	ret = bch2_mark_replicas(c, r);
-	if (ret)
-		goto err;
+	if (!c->sb.clean) {
+		/*
+		 * Mark journal replicas before we submit the write to guarantee
+		 * recovery will find the journal entries after a crash.
+		 *
+		 * If the filesystem is clean, we have to defer this until after
+		 * the write completes, so the filesystem isn't marked dirty
+		 * before anything is in the journal:
+		 */
+		struct bch_replicas_entry_v1 *r = &journal_seq_pin(j, le64_to_cpu(w->data->seq))->devs.e;
+		bch2_devlist_to_replicas(r, BCH_DATA_journal, w->devs_written);
+
+		ret = bch2_replicas_entry_get(c, r);
+		if (ret) {
+			r->nr_devs = 0;
+			goto err;
+		}
+	}
 
 	if (c->opts.nochanges)
 		goto no_io;
